@@ -1,18 +1,17 @@
-## Copyright (C) 2021  Brodie Gaslam
+## Copyright (C) Brodie Gaslam
 ##
 ## This file is part of "fansi - ANSI Control Sequence Aware String Functions"
 ##
 ## This program is free software: you can redistribute it and/or modify
 ## it under the terms of the GNU General Public License as published by
-## the Free Software Foundation, either version 2 of the License, or
-## (at your option) any later version.
+## the Free Software Foundation, either version 2 or 3 of the License.
 ##
 ## This program is distributed in the hope that it will be useful,
 ## but WITHOUT ANY WARRANTY; without even the implied warranty of
 ## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ## GNU General Public License for more details.
 ##
-## Go to <https://www.r-project.org/Licenses/GPL-2> for a copy of the license.
+## Go to <https://www.r-project.org/Licenses> for copies of the licenses.
 
 #' Replace Tabs With Spaces
 #'
@@ -27,8 +26,7 @@
 #'   `ctl` parameter only affects which _Control Sequences_ are considered zero
 #'   width.  Tabs will always be converted to spaces, irrespective of the `ctl`
 #'   setting.
-#' @seealso [fansi] for details on how _Control Sequences_ are
-#'   interpreted, particularly if you are getting unexpected results.
+#' @inherit has_ctl seealso
 #' @export
 #' @inheritParams substr_ctl
 #' @param x character vector or object coercible to character; any tabs therein
@@ -55,31 +53,17 @@
 #' ) )
 
 tabs_as_spaces <- function(
-  x, tab.stops=getOption('fansi.tab.stops'), warn=getOption('fansi.warn'),
-  ctl='all'
+  x, tab.stops=getOption('fansi.tab.stops', 8L),
+  warn=getOption('fansi.warn', TRUE), ctl='all'
 ) {
-  if(!is.character(x)) x <- as.character(x)
-  if(!is.logical(warn)) warn <- as.logical(warn)
-  if(length(warn) != 1L || is.na(warn))
-    stop("Argument `warn` must be TRUE or FALSE.")
-  if(!is.numeric(tab.stops) || !length(tab.stops) || any(tab.stops < 1))
-    stop("Argument `tab.stops` must be numeric and strictly positive")
-
-  if(!is.character(ctl))
-    stop("Argument `ctl` must be character.")
-  ctl.int <- integer()
-  if(length(ctl)) {
-    # duplicate values in `ctl` are okay, so save a call to `unique` here
-    if(anyNA(ctl.int <- match(ctl, VALID.CTL)))
-      stop(
-        "Argument `ctl` may contain only values in `",
-        deparse(VALID.CTL), "`"
-      )
-  }
-  term.cap.int <- seq_along(VALID.TERM.CAP)
+  ## modifies / creates NEW VARS in fun env
+  VAL_IN_ENV(
+    x=x, warn=warn, ctl=ctl, warn.mask=get_warn_worst(), tab.stops=tab.stops
+  )
+  term.cap.int <- 1L
   .Call(
-    FANSI_tabs_as_spaces, enc2utf8(x), as.integer(tab.stops), warn,
-    term.cap.int, ctl.int
+    FANSI_tabs_as_spaces, x, tab.stops, WARN.INT,
+    term.cap.int, CTL.INT
   )
 }
 #' Test Terminal Capabilities
@@ -99,7 +83,8 @@ tabs_as_spaces <- function(
 #' CSI SGR sequences are displayed as well.
 #'
 #' You should compare the screen output from this function to
-#' `getOption('fansi.term.cap')` to ensure that they are self consistent.
+#' `getOption('fansi.term.cap', dflt_term_cap)` to ensure that they are self
+#' consistent.
 #'
 #' By default `fansi` assumes terminals support bright and 256 color
 #' modes, and also tests for truecolor support via the $COLORTERM system
@@ -108,11 +93,10 @@ tabs_as_spaces <- function(
 #' Functions with the `term.cap` parameter like `substr_ctl` will warn if they
 #' encounter 256 or true color SGR sequences and `term.cap` indicates they are
 #' unsupported as such a terminal may misinterpret those sequences.  Bright
-#' codes in terminals that do not support them are more likely to be silently
-#' ignored, so `fansi` functions do not warn about those.
+#' codes and OSC hyperlinks in terminals that do not support them will likely be
+#' silently ignored, so `fansi` functions do not warn about those.
 #'
-#' @seealso [fansi] for details on how _Control Sequences_ are
-#'   interpreted, particularly if you are getting unexpected results.
+#' @seealso [`dflt_term_cap`], [`has_ctl`].
 #' @export
 #' @return character the test vector, invisibly
 #' @examples
@@ -170,21 +154,29 @@ fansi_lines <- function(txt, step=1) {
 #' Arbitrary text may contain characters with special meaning in HTML, which may
 #' cause HTML display to be corrupted if they are included unescaped in a web
 #' page.  This function escapes those special characters so they do not
-#' interfere with the HTML markup generated by e.g. [`sgr_to_html`].
+#' interfere with the HTML markup generated by e.g. [`to_html`].
 #'
 #' @export
 #' @family HTML functions
 #' @param x character vector
-#' @return `x`, but with "<", ">", "&", "'", and "\"" characters replaced by
-#'   their HTML entity codes, and Encoding set to UTF-8.
+#' @param what character(1) containing any combination of ASCII characters
+#'   "<", ">", "&", "'", or "\"".  These characters are special in HTML contexts
+#'   and will be substituted by their HTML entity code.  By default, all
+#'   special characters are escaped, but in many cases "<>&" or even "<>" might
+#'   be sufficient.
+#'  @return `x`, but with the `what` characters replaced by their HTML entity
+#'    codes.
+#' @note Non-ASCII strings are converted to and returned in UTF-8 encoding.
 #' @examples
 #' html_esc("day > night")
 #' html_esc("<SPAN>hello world</SPAN>")
 
-html_esc <- function(x) {
+html_esc <- function(x, what=getOption("fansi.html.esc", "<>&'\"")) {
   if(!is.character(x))
     stop("Argument `x` must be character, is ", typeof(x), ".")
-  .Call(FANSI_esc_html, enc2utf8(x))
+  if(!is.character(what))
+    stop("Argument `what` must be character, is ", typeof(what), ".")
+  .Call(FANSI_esc_html, enc_to_utf8(x), what)
 }
 
 #' Format Character Vector for Display as Code in HTML
@@ -217,21 +209,21 @@ html_code_block <- function(x, class='fansi-output') {
     "<PRE %s><CODE>%s</CODE></PRE>", class.all, paste0(x, collapse='\n')
   )
 }
-#' Set an Output Hook to Display ANSI CSI SGR in Rmarkdown
+#' Set an Output Hook Convert Control Sequences to HTML in Rmarkdown
 #'
 #' This is a convenience function designed for use within an `rmarkdown`
 #' document.  It overrides the `knitr` output hooks by using
-#' `knitr::knit_hooks$set`.  It replaces the hooks with ones that convert ANSI
-#' CSI SGR sequences into HTML.  In addition to replacing the hook functions,
+#' `knitr::knit_hooks$set`.  It replaces the hooks with ones that convert
+#' _Control Sequences_ into HTML.  In addition to replacing the hook functions,
 #' this will output a &lt;STYLE&gt; HTML block to stdout.  These two actions are
 #' side effects as a result of which R chunks in the `rmarkdown` document that
-#' contain ANSI CSI SGR are shown in their HTML equivalent form.
+#' contain CSI SGR are shown in their HTML equivalent form.
 #'
-#' The replacement hook function tests for the presence of ANSI CSI SGR
-#' sequences in chunk output with [`has_sgr`], and if it is detected then
+#' The replacement hook function tests for the presence of CSI SGR
+#' sequences in chunk output with [`has_ctl`], and if it is detected then
 #' processes it with the user provided `proc.fun`.  Chunks that do not contain
-#' ANSI CSI SGR are passed off to the previously set hook function.  The default
-#' `proc.fun` will run the output through [`html_esc`], [`sgr_to_html`], and
+#' CSI SGR are passed off to the previously set hook function.  The default
+#' `proc.fun` will run the output through [`html_esc`], [`to_html`], and
 #' finally [`html_code_block`].
 #'
 #' If you require more control than this function provides you can set the
@@ -243,7 +235,7 @@ html_code_block <- function(x, class='fansi-output') {
 #'   `rmarkdown`.
 #'
 #' @export
-#' @seealso [`has_sgr`], [`sgr_to_html`], [`html_esc`], [`html_code_block`],
+#' @seealso [`has_ctl`], [`to_html`], [`html_esc`], [`html_code_block`],
 #'   [`knitr` output hooks](https://yihui.org/knitr/hooks/#output-hooks),
 #'   [embedding CSS in
 #'   Rmd](https://bookdown.org/yihui/rmarkdown/language-engines.html#javascript-and-css),
@@ -256,7 +248,7 @@ html_code_block <- function(x, class='fansi-output') {
 #' @param class character the CSS class to give the output chunks.  Each type of
 #'   output chunk specified in `which` will be matched position-wise to the
 #'   classes specified here.  This vector should be the same length as `which`.
-#' @param proc.fun function that will be applied to output that contains ANSI
+#' @param proc.fun function that will be applied to output that contains
 #'   CSI SGR sequences.  Should accept parameters `x` and `class`, where `x` is
 #'   the output, and `class` is the CSS class that should be applied to
 #'   the &lt;PRE&gt;&lt;CODE&gt; blocks the output will be placed in.
@@ -266,13 +258,13 @@ html_code_block <- function(x, class='fansi-output') {
 #'   height 1.5 (as is the default setting in `rmarkdown` documents v1.1).
 #' @param split.nl TRUE or FALSE (default), set to TRUE to split input strings
 #'   by any newlines they may contain to avoid any newlines inside SPAN tags
-#'   created by [sgr_to_html()].  Some markdown->html renders can be configured
+#'   created by [to_html()].  Some markdown->html renders can be configured
 #'   to convert embedded newlines into line breaks, which may lead to a doubling
 #'   of line breaks.  With the default `proc.fun` the split strings are
 #'   recombined by [html_code_block()], but if you provide your own `proc.fun`
 #'   you'll need to account for the possibility that the character vector it
 #'   receives will have a different number of elements than the chunk output.
-#'   This argument only has an effect if chunk output contains ANSI CSI SGR
+#'   This argument only has an effect if chunk output contains CSI SGR
 #'   sequences.
 #' @param .test TRUE or FALSE, for internal testing use only.
 #' @return named list with the prior output hooks for each of `which`.
@@ -291,7 +283,7 @@ html_code_block <- function(x, class='fansi-output') {
 #' ## them (alternatively we could have done output as part of this call too)
 #'
 #' styles <- c(
-#'   getOption('fansi.style'),  # default style
+#'   getOption('fansi.style', dflt_css()),  # default style
 #'   "PRE.fansi CODE {background-color: transparent;}",
 #'   "PRE.fansi-error {background-color: #DD5555;}",
 #'   "PRE.fansi-warning {background-color: #DDDD55;}",
@@ -316,9 +308,9 @@ html_code_block <- function(x, class='fansi-output') {
 set_knit_hooks <- function(
   hooks, which='output',
   proc.fun=function(x, class)
-    html_code_block(sgr_to_html(html_esc(x)), class=class),
+    html_code_block(to_html(html_esc(x)), class=class),
   class=sprintf("fansi fansi-%s", which),
-  style=getOption("fansi.css"),
+  style=getOption("fansi.css", dflt_css()),
   split.nl=FALSE,
   .test=FALSE
 ) {
@@ -377,7 +369,7 @@ set_knit_hooks <- function(
       # If the output has SGR in it, then convert to HTML and wrap
       # in PRE/CODE tags
 
-      if(any(has_sgr(x))) {
+      if(any(has_ctl(x, c('sgr', 'url')))) {
         if(split.nl) x <- unlist(strsplit_sgr(x, '\n', fixed=TRUE))
         res <- try(proc.fun(x=x, class=class))
         if(inherits(res, "try-error"))
@@ -426,7 +418,7 @@ set_knit_hooks <- function(
   if(.test) list(old.hooks=old.hook.list, new.hooks=new.hook.list, res=set.res)
   else old.hook.list
 }
-#' Show 8 Bit ANSI CSI SGR Colors
+#' Show 8 Bit CSI SGR Colors
 #'
 #' Generates text with each 8 bit SGR code (e.g. the "###" in "38;5;###") with
 #' the background colored by itself, and the foreground in a contrasting color
@@ -471,3 +463,77 @@ sgr_256 <- function() {
     gs1, gs2
   )
 }
+
+# To test growable buffer.
+
+size_buff <- function(x) .Call(FANSI_size_buff, x)
+size_buff_prot_test <- function() {
+  raw <- .Call(FANSI_size_buff_prot_test)
+  res <- raw[-1L]
+  names(res) <- c('n', 'prev', 'self')
+  res <- as.data.frame(res)
+  # stringsAsFactors issues
+  res[['prev']] <- as.character(res[['prev']])
+  res[['self']] <- as.character(res[['self']])
+  rownames(res) <- raw[[1L]]
+  # remap the addresses so they are consistent across different runs
+  addresses <- do.call(rbind, res[c('prev', 'self')])
+  res[['prev']] <- match(res[['prev']], addresses)
+  res[['self']] <- match(res[['self']], addresses)
+  res
+}
+#' Display Strings to Terminal
+#'
+#' Shortcut for [`writeLines`] with an additional terminating "ESC&#91;0m".
+#'
+#' @keywords internal
+#' @export
+#' @param ... character vectors to display.
+#' @param end character what to output after the primary inputs.
+#' @return whatever writeLines returns
+
+fwl <- function(..., end='<END>\033[0m') {
+  writeLines(c(..., end))
+}
+
+#' Default Arg Helper Funs
+#'
+#' Terminal capabilities are assumed to include bright and 256 color SGR codes.
+#' 24 bit color support is detected based on the `COLORTERM` environment
+#' variable.
+#'
+#' Default CSS may exceed or fail to cover the interline distance when two lines
+#' have background colors.  To ensure lines are exactly touching use
+#' inline-block, although that has its own issues.  Otherwise specify your own
+#' CSS.
+#'
+#' @seealso [`term_cap_test`].
+#' @export
+#' @return character to use as default value for `fansi` parameter.
+
+dflt_term_cap <- function() {
+  c(
+    if(isTRUE(Sys.getenv('COLORTERM') %in% c('truecolor', '24bit')))
+    'truecolor',
+    'bright', '256'
+  )
+}
+#' @rdname dflt_term_cap
+#' @export
+
+dflt_css <- function() {
+  "PRE.fansi SPAN {padding-top: .25em; padding-bottom: .25em};"
+}
+#' Report Unicode Database Version In-use
+#'
+#' Corresponds to the database used to generate character display width look-up
+#' tables by `fansi`.
+#'
+#' @export
+#' @return a string of the Unicode version.
+#' @examples
+#' fansi_unicode_version()
+
+fansi_unicode_version <- function() {
+  .Call(FANSI_unicode_version)
+}
